/***********************************************
 * Real Time Clock State reporter
 *  
 *   Copyright (C) 1999  Kenichi Okuyama
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <Pilot.h>
#include <stdio.h>
#include <string.h>


#define testDBCreator	'RTCs'
#define testDBType	'Data'
#define testDBname	"RTC-log"
#define HEADERTEXT	"\nOKKYDOKKY-OKKYDOKKY-OKKYDOKKY\nPlease remove everyline above.\n"


/*
 * MC68000 CPU bitmasks
 */
#define CPUSR_TRACEBIT		0x8000
#define CPUSR_SUPERVISORBIT	0x2000
#define CPUSR_INTRMASK		0x0700
#define CPUSR_EXTEND		0x0010
#define CPUSR_NEGATIVE		0x0008
#define CPUSR_ZERO		0x0004
#define CPUSR_OVERFLOW		0x0002
#define CPUSR_CARRY		0x0001

/*
 * Chip and Mask ID Address
 * CHIPANDMASKADDR      : Chip and Mask r/o register.
 * SOFTWAREIDADDR       : S/W ID register.
 */
#define CHIPANDMASKADDR (0xfffff004)
#define SOFTWAREIDADDR  (0xfffff006)


/*
 * Interrupt Handling registers
 * IMR_ADDR     : Interrupt Mask Registers.
 * ISR_ADDR     : Interrupt Status Registers.
 * IPR_ADDR     : Interrupt Pending Registers.
 * ISR only holds interrupt that is not being masked.
 * IPR only holds interrupt that is being masked.
 * ISR = (Real Interrupt) & IMR;
 * ISR = (Real Interrupt) & (not IMR);
 */
#define IMR_ADDR        (0xfffff304)
#define ISR_ADDR        (0xfffff30c)
#define IPR_ADDR        (0xfffff310)

/*
 * RealTimeClock memory mapped registers address for Dragonball.
 * CLOCKADDR    : real-time clock register.
 * ALARMADDR    : real-time Alarm register.
 * CNTR_ADDR    : real-time clock control bit fields.
 * INTRSADDR    : interruption state register.
 * INTREADDR    : interruption enable/disable register.
 */
#define RTC_CLOCKADDR   (0xfffffb00UL)
#define RTC_ALARMADDR   (0xfffffb04UL)
#define RTC_WATCHADDR   (0xfffffb0aUL)
#define RTC_CNTR_ADDR   (0xfffffb0cUL)
#define RTC_INTRSADDR   (0xfffffb0eUL)
#define RTC_INTREADDR   (0xfffffb10UL)
#define RTC_SWADDR      (0xfffffb12UL)
#define RTC_DAYCTADDR	(0xfffffb1aUL)
#define RTC_DAYALADDR	(0xfffffbacUL)

typedef struct {
    unsigned long       mask;
    unsigned long       status;
    unsigned long       pending;
} intrreg;

typedef struct {
    unsigned char       hour;
    unsigned char       minute;
    unsigned char       second;
} realtime;

typedef struct {
    unsigned long	clock;
    unsigned long       alarm;
    unsigned short      watchdog;
    unsigned short      control;
    unsigned short      intr_state;
    unsigned short      intr_mask;
    unsigned short      stopwatch;
    unsigned short      daycount;
    unsigned short      dayalarm;
    int                 isEZ;
} rtc_states;

typedef struct {
    long		year;
    unsigned char	month;
    unsigned char       day;
    unsigned char	hour;
    unsigned char	minute;
    unsigned char	second;
} os_status;

typedef struct {
    unsigned short      csa0;   /* 0xfffff100 *//***/
    unsigned short      csa1;   /* 0xfffff102 *//***/
    unsigned short      csa2;   /* 0xfffff104 *//***/
    unsigned short      csa3;   /* 0xfffff106 *//***/
    unsigned short      gmska;  /* 0xfffff108 */
    unsigned short      gmskb;  /* 0xfffff10a */
    unsigned short      gmskc;  /* 0xfffff10c */
    unsigned short      gmskd;  /* 0xfffff10e */
    unsigned long       csoa0;  /* 0xfffff110 *//***/
    unsigned long       csoa1;  /* 0xfffff114 *//***/
    unsigned long       csoa2;  /* 0xfffff118 *//***/
    unsigned long       csoa3;  /* 0xfffff11c */
    unsigned long       csob0;  /* 0xfffff120 */
    unsigned long       csob1;  /* 0xfffff124 */
    unsigned long       csob2;  /* 0xfffff128 */
    unsigned long       csob3;  /* 0xfffff12c */
    unsigned long       csoc0;  /* 0xfffff130 */
    unsigned long       csoc1;  /* 0xfffff134 */
    unsigned long       csoc2;  /* 0xfffff138 */
    unsigned long       csoc3;  /* 0xfffff13c */
    unsigned long       csod0;  /* 0xfffff140 */
    unsigned long       csod1;  /* 0xfffff144 */
    unsigned long       csod2;  /* 0xfffff148 */
    unsigned long       csod3;  /* 0xfffff14c */
} chip_select_status;


DWord		PilotMain( Word cmd, Ptr cmdPBP, Word launchFlags );
static void		WriteCPUStatusReg( DmOpenRef logDB, Word cmd );
static unsigned short   GetCPUStatusReg( void );
static unsigned long    GetCPUInfo( char *bufP );
static void		WriteInterRegs( DmOpenRef logDB, rtc_states *rtcsP );
static void             WriteChipSelectInfo( DmOpenRef logDB, rtc_states *rtcsP );
static void             WriteTimeInfos( DmOpenRef logDB, rtc_states *rtcsP );
static void             GetTimeOfOS( os_status *ossP );
static void             GetTimeOfUnit( rtc_states *rtcsP );
static void             Log( DmOpenRef logDB, char *strP );


DWord
PilotMain( Word cmd, Ptr cmdPBP, Word launchFlags )
{
    rtc_states          rtcs;
    LocalID		dbID;
    DmOpenRef   	logDB;
    unsigned long	chipval;
    unsigned char	chipid, maskid;

    dbID        = DmFindDatabase( 0, testDBname );
    if ( dbID == 0 ) {
        char    chipname[32];
        char    buffer[128];

        if ( DmCreateDatabase( 0,
                               testDBname,
                               testDBCreator,
                               testDBType, false )) {
            return 1;
        }

        chipval	= GetCPUInfo( chipname );
        chipid	= (unsigned char)(( chipval >> 24 ) & 0x00ff );
        maskid	= (unsigned char)(( chipval >> 16 ) & 0x00ff );

        dbID        = DmFindDatabase( 0, testDBname );
        if ( dbID == 0 ) {
            return 1;
        }
        logDB       = DmOpenDatabase( 0, dbID, dmModeReadWrite );
        sprintf( buffer,
                 HEADERTEXT
                 "CHIPID = 0x%x  MASKID = 0x%x S/WID = 0x%x\n"
                 "chiptype = %s\n",
                 chipid, maskid,
                 (unsigned int)( chipval & 0xffff ),
                 chipname );
        Log( logDB, buffer );
    } else {
        logDB       = DmOpenDatabase( 0, dbID, dmModeReadWrite );
        if ( logDB == 0 ) {
            return 1;
        }
        chipval	= GetCPUInfo( NULL );
        chipid	= (unsigned char)(( chipval >> 24 ) & 0x00ff );
        maskid	= (unsigned char)(( chipval >> 16 ) & 0x00ff );
    }

    switch( chipid ) {
    case 0x1c:
    case 0x33:
        rtcs.isEZ	= 0;
        break;
    case 0x00:
    case 0x45:
        rtcs.isEZ	= 1;
        break;
    default:
        /* it's not that this is Dragonball */
        /* this only means it's not listed in Chip list */
        rtcs.isEZ	= 0;
        break;
    }

    WriteCPUStatusReg( logDB, cmd );
    WriteInterRegs( logDB, &rtcs );
    WriteTimeInfos( logDB, &rtcs );
    WriteChipSelectInfo( logDB, &rtcs );
    DmCloseDatabase( logDB );
    return 0;
}


static void
WriteCPUStatusReg( DmOpenRef logDB, Word cmd )
{
    char                buffer[64];
    unsigned short	status_reg;

    status_reg  = GetCPUStatusReg();
    sprintf( buffer,
             "\n"
             "Boot code: %d\n"
             "CPU_state(0x%x)[%s%s INTR:%d %s%s%s%s%s]\n",
             cmd,
             status_reg,
             ((status_reg & CPUSR_TRACEBIT)? "T" : " " ),
             ((status_reg & CPUSR_SUPERVISORBIT) ? "S" : "U"),
             (status_reg & CPUSR_INTRMASK ),
             ((status_reg & CPUSR_EXTEND) ? "E" : " "),
             ((status_reg & CPUSR_NEGATIVE) ? "N" : " "),
             ((status_reg & CPUSR_ZERO) ? "Z" : " "),
             ((status_reg & CPUSR_OVERFLOW) ? "O" : " "),
             ((status_reg & CPUSR_CARRY) ? "C" : " "));
    Log( logDB, buffer );
}

static unsigned short
GetCPUStatusReg( void )
{
    unsigned long      regval;

    asm ( "move %%sr, %0" : "=d" (regval) );
    return regval;
}

static unsigned long
GetCPUInfo( char *bufP )
{
    unsigned long       retval;
    unsigned short      chipmask, software;

    chipmask    = *((volatile unsigned short *)(CHIPANDMASKADDR));
    software	= *((volatile unsigned short *)(SOFTWAREIDADDR));
    retval	= (((unsigned long)(chipmask)) << 16 )
        | (((unsigned long)(software)) << 0 );

    if ( bufP ) {
        switch(( chipmask >> 8 )&0xff ) {
        case 0x1c:
            sprintf( bufP, "MC68328 3G58E/0H51K" );
            break;

        case 0x33:
            switch( chipmask & 0xff ) {
            case 0x30:
                sprintf( bufP, "MC68328 1H58B/0H38T" );
                break;
            default:
                sprintf( bufP, "?MC68328? unknown" );
                break;
            }
            break;
        case 0x00:
            switch( chipmask & 0xff ) {
            case 0x01:
                sprintf( bufP, "MC68EZ328 0H31J" );
                break;
            default:
                sprintf( bufP, "?MC68EZ328? unknown" );
                break;
            }
            break;
        case 0x45:
            switch( chipmask & 0xff ) {
            case 0x02:
                sprintf( bufP, "MC68EZ328 0J75C" );
                break;
            case 0x05:
                sprintf( bufP, "MC68EZ328 1J83G" );
                break;
            default:
                sprintf( bufP, "?MC68EZ328? unknown" );
                break;
            }
            break;
        default:
            sprintf( bufP, "unknown" );
            break;
        }
    }

    return retval;
}

static void
WriteInterRegs( DmOpenRef logDB, rtc_states *rtcsP )
{
    char        buffer[64];
    intrreg     iregs;

    iregs.mask		= *((volatile unsigned long *)(IMR_ADDR));
    iregs.status	= *((volatile unsigned long *)(ISR_ADDR));
    iregs.pending	= *((volatile unsigned long *)(IPR_ADDR));

    sprintf( buffer,
             "IntrReg( mask:0x%08lx, status:0x%08lx, pend:0x%08lx )\n",
             iregs.mask, iregs.status, iregs.pending );
    Log( logDB, buffer );
}


static void
WriteChipSelectInfo( DmOpenRef logDB, rtc_states *rtcsP )
{
    chip_select_status  css;
    char        	buffer[256];

    css.csa0  = *((volatile unsigned short *)(0xfffff100));
    css.csa1  = *((volatile unsigned short *)(0xfffff102));
    css.csa2  = *((volatile unsigned short *)(0xfffff104));
    css.csa3  = *((volatile unsigned short *)(0xfffff106));

    css.gmska = *((volatile unsigned short *)(0xfffff108));
    css.gmskb = *((volatile unsigned short *)(0xfffff10a));
    css.gmskc = *((volatile unsigned short *)(0xfffff10c));
    css.gmskd = *((volatile unsigned short *)(0xfffff10e));

    css.csoa0 = *((volatile unsigned long  *)(0xfffff110));
    css.csoa1 = *((volatile unsigned long  *)(0xfffff114));
    css.csoa2 = *((volatile unsigned long  *)(0xfffff118));
    css.csoa3 = *((volatile unsigned long  *)(0xfffff11c));

    css.csob0 = *((volatile unsigned long  *)(0xfffff120));
    css.csob1 = *((volatile unsigned long  *)(0xfffff124));
    css.csob2 = *((volatile unsigned long  *)(0xfffff128));
    css.csob3 = *((volatile unsigned long  *)(0xfffff12c));

    css.csoc0 = *((volatile unsigned long  *)(0xfffff130));
    css.csoc1 = *((volatile unsigned long  *)(0xfffff134));
    css.csoc2 = *((volatile unsigned long  *)(0xfffff138));
    css.csoc3 = *((volatile unsigned long  *)(0xfffff13c));

    css.csod0 = *((volatile unsigned long  *)(0xfffff140));
    css.csod1 = *((volatile unsigned long  *)(0xfffff144));
    css.csod2 = *((volatile unsigned long  *)(0xfffff148));
    css.csod3 = *((volatile unsigned long  *)(0xfffff14c));

    sprintf( buffer,
             "chipSelect(0x%04x, 0x%04x, 0x%04x, 0x%04x,"
                        "0x%04x, 0x%04x, 0x%04x, 0x%04x\n"
             "        a  0x%08lx, 0x%08lx, 0x%08lx, 0x%08lx\n"
             "        b  0x%08lx, 0x%08lx, 0x%08lx, 0x%08lx\n"
             "        c  0x%08lx, 0x%08lx, 0x%08lx, 0x%08lx\n"
             "        d  0x%08lx, 0x%08lx, 0x%08lx, 0x%08lx )\n",
             css.csa0,
             css.csa1,
             css.csa2,
             css.csa3,
             css.gmska,
             css.gmskb,
             css.gmskc,
             css.gmskd,
             css.csoa0,
             css.csoa1,
             css.csoa2,
             css.csoa3,
             css.csob0,
             css.csob1,
             css.csob2,
             css.csob3,
             css.csoc0,
             css.csoc1,
             css.csoc2,
             css.csoc3,
             css.csod0,
             css.csod1,
             css.csod2,
             css.csod3 );

    Log( logDB, buffer );
}


static void
WriteTimeInfos( DmOpenRef logDB, rtc_states *rtcsP )
{
    os_status   oss;
    char        buffer[256];

    GetTimeOfOS( &oss );
    GetTimeOfUnit( rtcsP );

    if ( rtcsP->isEZ ) {
        sprintf( buffer,
                 "rtc( clock(%3d %2d:%2d:%2d), alarm(%3d %2d:%2d:%2d), "
                 "control0x%04x, intrst0x%04x, intrmsk0x%04x, "
                 "sw0x%04x, watchdog0x%04x )\n"
                 "os( %4ld/%2d/%2d %2d:%2d:%2d )\n",
                 (int)(( rtcsP->daycount ) & 0x01ff ),
                 (int)(( rtcsP->clock >> 24 ) & 0x00ff ),
                 (int)(( rtcsP->clock >> 16 ) & 0x00ff ),
                 (int)(( rtcsP->clock >>  0 ) & 0x00ff ),
                 (int)(( rtcsP->dayalarm ) & 0x01ff ),
                 (int)(( rtcsP->alarm>> 24 ) & 0x00ff ),
                 (int)(( rtcsP->alarm >> 16 ) & 0x00ff ),
                 (int)(( rtcsP->alarm >>  0 ) & 0x00ff ),
                 rtcsP->control,
                 rtcsP->intr_state,
                 rtcsP->intr_mask,
                 rtcsP->stopwatch,
                 rtcsP->watchdog,
                 oss.year, oss.month, oss.day,
                 oss.hour, oss.minute, oss.second );
    } else {
        sprintf( buffer,
                 "rtc( clock(%3d %2d:%2d:%2d), alarm(%3d %2d:%2d:%2d), "
                 "control0x%04x, intrst0x%04x, intrmsk0x%04x, "
                 "sw0x%04x, watchdog0x%04x )\n"
                 "os( %4ld/%2d/%2d %2d:%2d:%2d )\n",
                 (int)(( rtcsP->daycount ) & 0x01ff ),
                 (int)(( rtcsP->clock >> 24 ) & 0x00ff ),
                 (int)(( rtcsP->clock >> 16 ) & 0x00ff ),
                 (int)(( rtcsP->clock >>  0 ) & 0x00ff ),
                 (int)(( rtcsP->dayalarm ) & 0x01ff ),
                 (int)(( rtcsP->alarm>> 24 ) & 0x00ff ),
                 (int)(( rtcsP->alarm >> 16 ) & 0x00ff ),
                 (int)(( rtcsP->alarm >>  0 ) & 0x00ff ),
                 rtcsP->control,
                 rtcsP->intr_state,
                 rtcsP->intr_mask,
                 rtcsP->stopwatch,
                 rtcsP->watchdog,
                 oss.year, oss.month, oss.day,
                 oss.hour, oss.minute, oss.second );
    }

    Log( logDB, buffer );
}


static void
GetTimeOfOS( os_status *ossP )
{
    ULong		curTime;
    DateTimeType        dateTime;

    curTime	= TimGetSeconds();
    TimSecondsToDateTime( curTime, &dateTime );
    ossP->year		= (unsigned long)(dateTime.year);
    ossP->month		= dateTime.month;
    ossP->day		= dateTime.day;
    ossP->hour		= dateTime.hour;
    ossP->minute	= dateTime.minute;
    ossP->second	= dateTime.second;
}

static void
GetTimeOfUnit( rtc_states *rtcsP )
{

    /* get rct Day alarm register */
    if ( rtcsP->isEZ ) {
        rtcsP->dayalarm	= *((volatile unsigned short *)(RTC_DAYALADDR));
    }

    /* get rct control register */
    rtcsP->control	= *((volatile unsigned short *)(RTC_CNTR_ADDR));

    /* get interrupt enable */
    rtcsP->intr_mask	= *((volatile unsigned short *)(RTC_INTREADDR));

    /* get alarm register */
    rtcsP->alarm	= *((volatile unsigned long *)(RTC_ALARMADDR));

    /* get watchdog register */
    if ( rtcsP->isEZ ) {
        rtcsP->watchdog	= *((volatile unsigned short *)(RTC_WATCHADDR));
    }

    /* get StopWatch register */
    rtcsP->stopwatch	= *((volatile unsigned short *)(RTC_SWADDR));

    /* get RTC interrupt status register */
    rtcsP->intr_state	= *((volatile unsigned short *)(RTC_INTRSADDR));

    /* get rct Day count register */
    if ( rtcsP->isEZ ) {
        rtcsP->daycount	= *((volatile unsigned short *)(RTC_DAYCTADDR));
        if ( rtcsP->daycount != *((volatile unsigned short *)(RTC_DAYCTADDR)) ) {
            rtcsP->daycount	= *((volatile unsigned short *)(RTC_DAYCTADDR));
        }
    }

    /* get current Unit time */
    rtcsP->clock	= *((volatile unsigned long *)(RTC_CLOCKADDR));
    if ( rtcsP->clock != *((volatile unsigned long *)(RTC_CLOCKADDR)) ) {
        rtcsP->daycount	= *((volatile unsigned short *)(RTC_DAYCTADDR));
        rtcsP->clock	= *((volatile unsigned long *)(RTC_CLOCKADDR));
    }
}



static void
Log( DmOpenRef logDB, char *strP )
{
    VoidHand	recH;
    char	*recP;
    int         len;
    UInt	index;

    len = strlen( strP );
    index	= DmNumRecords( logDB );
    recH	= DmNewRecord( logDB, &index, len );
    if ( recH != 0 ) {
        recP	= MemHandleLock( recH );
        DmWrite( recP, 0, strP, len );
        MemHandleUnlock( recH );
        DmReleaseRecord( logDB, index, true );
    }
}
